from typing import Optional, Any, Dict
import logging
import socket
import threading
import time

logger = logging.getLogger(__name__)


class RobotiqGripper:
    STATUS_RESET = 0
    STATUS_ACTIVATING = 1
    STATUS_ACTIVE = 3

    def __init__(self, host: str, base_port: int = 63352, id: int = 1, timeout: float = 2.0, connect: bool = True):
        self.logger = logger.getChild(self.__class__.__name__)
        self.host = host
        self.port = base_port + id - 1  # convert gripper id to port number, id 1 starts at 63352
        self.timeout = timeout
        self.connection: Optional[socket.socket] = None

        if connect:
            self.connect()

    @property
    def connected(self) -> bool:
        return self.connection is not None

    @property
    def status(self) -> int:
        return int(self.get_register('STA'))

    @property
    def active(self) -> bool:
        return self.status == self.STATUS_ACTIVE

    @property
    def position_int(self) -> int:
        return int(self.get_register('POS'))

    @property
    def position(self) -> float:
        return self.position_int / 255.0

    def connect(self):
        if self.connection:
            return

        self.logger.debug(f'Connecting to Robotiq gripper at {self.host}:{self.port}')
        self.connection = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connection.connect((self.host, self.port))
        self.connection.settimeout(self.timeout)
        self.logger.info(f'Connected to Robotiq gripper at {self.host}:{self.port}')

    def disconnect(self):
        self.logger.info(f'Disconnecting from Robotiq gripper at {self.host}:{self.port}')
        self.connection.close()
        self.connection = None

    def request(self, request: bytes) -> bytes:
        self.logger.debug(f'Sending gripper request: {request}')
        self.connection.send(request + b'\n')
        response = self.connection.recv(1024)
        self.logger.debug(f'Gripper response: {response}')
        return response

    def set_registers(self, registers: Dict[str, Any]):
        self.logger.debug(f'Setting grippper registers {registers}')
        command = 'SET ' + ' '.join([f'{name} {val}' for name, val in registers.items()])
        response = self.request(command.encode())

        if response != b'ack':
            raise RobotiqGripperSetError(f'Error setting registers, no ack received (response: {response})')

    def get_register(self, name: str) -> bytes:
        response = self.request(f'GET {name}'.encode())

        if not response.startswith(f'{name} '.encode()):
            raise RobotiqGripperInvalidResponseError(f'Invalid response: {response}')

        return response[len(name) + 1:].strip()

    def activate(self, wait: bool = True, timeout: float = 2.0, force: bool = False):
        if self.active and not force:
            return

        # clear and set activation register (resets faults)
        self.set_registers({'ACT': 0})
        self.set_registers({'ACT': 1})

        if not wait:
            return

        start_time = time.time()

        while time.time() - start_time < timeout:
            if self.active:
                return

            time.sleep(0.01)

        raise RobotiqGripperTimeoutError(f'Timeout while waiting for gripper to activate')

    def move(self, position: float, force: float = 0.5, velocity: float = 0.5, wait: bool = True, timeout: float = 10.0):
        if position < 0 or position > 1:
            raise RobotiqGripperInvalidPositionError(f'Invalid position, must be float between 0 and 1: {position}')

        if force < 0 or force > 1:
            raise RobotiqGripperInvalidForceError(f'Invalid force, must be float between 0 and 1: {force}')

        if velocity < 0 or velocity > 1:
            raise RobotiqGripperInvalidSpeedError(f'Invalid speed, must be float between 0 and 1: {force}')

        position_int = round(position * 255)
        force_int = round(force * 255)
        velocity_int = round(velocity * 255)

        self.set_registers({'POS': position_int, 'SPE': velocity_int, 'FOR': force_int, 'GTO': 1})

        if wait:
            time.sleep(0.1)
            self.wait_for_stop(timeout)

    def wait_for_stop(self, timeout: float = 10.0):
        last_pos = self.position_int
        start_time = time.time()

        while time.time() - start_time < timeout:
            time.sleep(0.1)
            current_pos = self.position_int

            if current_pos == last_pos:
                return

            last_pos = current_pos

        raise RobotiqGripperTimeoutError(f'Timeout while waiting for gripper to stop')

    def open(self, force: Optional[float] = None, velocity: Optional[float] = None, wait: bool = True, timeout: float = 10.0):
        self.move(0, force=force, velocity=velocity, wait=wait, timeout=timeout)

    def close(self, force: Optional[float] = None, velocity: Optional[float] = None, wait: bool = True, timeout: float = 10.0):
        self.move(1, force=force, velocity=velocity, wait=wait, timeout=timeout)


class RobotiqGripperError(Exception):
    pass


class RobotiqGripperSetError(RobotiqGripperError):
    pass


class RobotiqGripperConnectionError(RobotiqGripperError):
    pass


class RobotiqGripperInvalidResponseError(RobotiqGripperError):
    pass


class RobotiqGripperInvalidForceError(RobotiqGripperError):
    pass


class RobotiqGripperInvalidSpeedError(RobotiqGripperError):
    pass


class RobotiqGripperInvalidPositionError(RobotiqGripperError):
    pass


class RobotiqGripperTimeoutError(RobotiqGripperError):
    pass


from typing import Optional, List, Union
import time
import math
import urx
from urx import ursecmon
import math3d
from hein_robots.robotics import Location, Twist, Wrench, Units, Frame
from hein_robots.robotiq.gripper import RobotiqGripper
from hein_robots.base import robot_arms


class UR3Arm(robot_arms.RobotArm):
    def __init__(self, host: str = '192.168.0.100', default_velocity: float = 250, max_velocity: float = 500, position_units: str = Units.MILLIMETERS,
                 default_joint_velocity: float = 20.0, max_joint_velocity = 180.0, gripper_default_velocity: float = 0.5, gripper_default_force: float = 0.5,
                 gripper_id: int = 1, connect: bool = True):
        super().__init__(default_velocity=default_velocity, max_velocity=max_velocity, position_units=position_units,
                         gripper_default_velocity=gripper_default_velocity, gripper_default_force=gripper_default_force)
        self._default_joint_velocity = default_joint_velocity
        self.max_joint_velocity = max_joint_velocity
        self.host = host
        self._robot: Optional[urx.Robot] = None
        self.gripper = RobotiqGripper(self.host, id=gripper_id, connect=False)

        if connect:
            self.connect()

    @property
    def robot(self) -> urx.Robot:
        if self._robot is None:
            raise robot_arms.RobotArmNotConnectedError(f'UR3 arm not connected, try running connect()')

        return self._robot

    @property
    def connected(self) -> bool:
        return self._robot is not None

    @property
    def default_joint_velocity(self) -> float:
        return self._default_velocity

    @default_joint_velocity.setter
    def default_joint_velocity(self, value: float):
        if value < 0 or value > self.max_joint_velocity:
            raise robot_arms.RobotArmInvalidVelocityError(f'Invalid velocity: {value} degrees/s, must be less than {self.max_joint_velocity} degrees/s')

        self._default_velocity = value

    @property
    def acceleration(self) -> float:
        return 0.0

    @property
    def velocity(self) -> float:
        return 0.0

    @property
    def pose(self):
        return self.robot.get_pose()

    @property
    def location(self) -> Location:
        return Location.from_matrix(self.pose.array).convert_m_to_mm()

    @property
    def joint_positions(self) -> List[float]:
        return [math.degrees(position) for position in self.robot.getj()]

    @property
    def joint_count(self) -> int:
        return len(self.joint_positions)

    @property
    def tool_offset(self) -> Location:
        raise robot_arms.RobotArmNotSupportedError(f'Reading UR3Arm tool_offset not supported')

    @tool_offset.setter
    def tool_offset(self, value: Location):
        self.robot.set_tcp(*value.position, *value.orientation)

    @property
    def tool_mass(self) -> float:
        raise robot_arms.RobotArmNotSupportedError(f'Reading UR3Arm tool_mass not supported')

    @tool_mass.setter
    def tool_mass(self, value: float):
        self.robot.set_payload(value)

    @property
    def gripper_position(self) -> float:
        return self.gripper.position

    def connect(self, retries=5):
        while retries > 0:
            try:
                self._robot = urx.Robot(self.host)
                self.gripper.connect()
                return
            except ursecmon.TimeoutException:
                print('Retrying')
                time.sleep(1)
                retries -= 1

        raise robot_arms.RobotArmConnectionError(f'Cannot connect to UR3 at host {self.host}')

    def disconnect(self):
        self._robot.close()
        self._robot = None
        self.gripper.disconnect()

    def stop(self):
        self.robot.stop()

    def close(self):
        self.robot.close()

    def emergency_stop(self):
        self.stop()

    def wait(self, timeout: Optional[float] = None, wait_for_running: bool = False):
        start_time = time.time()

        if wait_for_running:
            while not self.robot.is_program_running():
                if timeout is not None and (time.time() - start_time) > timeout:
                    raise robot_arms.RobotArmWaitTimeoutError(f'Timeout while waiting for UR3 arm to start')

                time.sleep(0.01)

        while self.robot.is_program_running():
            if timeout is not None and (time.time() - start_time) > timeout:
                raise robot_arms.RobotArmWaitTimeoutError(f'Timeout while waiting for UR3 arm to stop')

            time.sleep(0.01)

    def move_to_location(self, location: Location, frame: Optional[Frame] = None,
                         velocity: Optional[float] = None, acceleration: Optional[float] = None,
                         relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        vel = velocity if velocity is not None else self._default_velocity
        acc = acceleration if acceleration is not None else vel * 2
        # use the reference frame (or an identity frame) to transform the relative location to a location in the base reference frame
        relative_position = (frame or Location()) * location
        transform = self.robot.csys * math3d.Transform(relative_position.convert_mm_to_m().matrix)
        distance = self.robot._get_lin_dist(transform.pose_vector)

        self.robot.movel(transform.pose_vector, acc=acc / 1000, vel=vel / 1000, relative=relative, wait=False)

        # don't wait if we aren't going to move
        if wait and distance > 1e-3:
            self.wait(timeout, wait_for_running=True)

    def move_joints(self, joint_positions: List[float],
                    velocity: Optional[float] = None, acceleration: Optional[float] = None,
                    relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
        if len(joint_positions) != self.joint_count:
            raise robot_arms.RobotArmInvalidJointsError(f'Invalid number of UR3 arm joints, must be {self.joint_count}')

        vel = velocity if velocity is not None else self._default_joint_velocity
        acc = acceleration if acceleration is not None else vel * 2
        joint_radians = [math.radians(position) for position in joint_positions]
        distance = self.robot._get_joints_dist(joint_radians)

        try:
            self.robot.movej(joint_radians, vel=math.radians(vel), acc=math.radians(acc), relative=relative, wait=False)
        except urx.urrobot.RobotException:
            pass

        # don't wait if we aren't going to move
        if wait and distance > 1e-3:
            self.wait(timeout, wait_for_running=True)

    def move_circular(self, midpoint: Location, end: Location, frame: Optional[Frame] = None,
                      velocity: Optional[float] = None, acceleration: Optional[float] = None,
                      wait: bool = True, timeout: Optional[float] = None):
        vel = velocity if velocity is not None else self._default_velocity
        acc = acceleration if acceleration is not None else vel * 2
        # use the reference frame (or an identity frame) to transform the relative location to a location in the base reference frame
        relative_midpoint = (frame or Location()) * midpoint
        relative_end = (frame or Location()) * end
        transform_midpoint = self.robot.csys * math3d.Transform(relative_midpoint.convert_mm_to_m().matrix)
        transform_end = self.robot.csys * math3d.Transform(relative_end.convert_mm_to_m().matrix)

        self.robot.movec(transform_midpoint.pose_vector, transform_end.pose_vector, acc=acc / 1000, vel=vel / 1000, wait=False)

        if wait:
            self.wait(timeout, wait_for_running=True)

    def wait_for_gripper_stop(self, timeout: Optional[float] = None):
        self.gripper.wait_for_stop(timeout=timeout)

    def open_gripper(self, position: Optional[Union[float, bool]] = None, force: Optional[float] = None, velocity: Optional[float] = None,
                     wait: bool = True, timeout: Optional[float] = None):
        if position is None:
            position = 0.0
        elif isinstance(position, bool):
            position = 0.0 if position else 1.0

        self.gripper.move(position, force=force or self.gripper_default_force, velocity=velocity or self.gripper_default_velocity, wait=wait, timeout=timeout or 10.0)


from typing import Union, Tuple, Dict, List, Optional, Any, Iterable
import numpy as np
from hein_robots.robotics import Location, Cartesian


GridSpacingType = Union[Tuple[float, float], Union[float, float, float], Cartesian]


class LocationGroup:
    def __len__(self):
        return 0

    def __getitem__(self, item):
        return None

    def indexes(self) -> List[Any]:
        return []


class BaseGrid(LocationGroup):
    def __init__(self, rows: int, columns: int):
        self.rows = rows
        self.columns = columns
        self.location_grid: Optional[np.ndarray] = None

    def __len__(self):
        return len(self.location_grid)

    def __getitem__(self, item: Union[Tuple[int, int], int, str]) -> Union[Location, np.ndarray]:
        if isinstance(item, tuple):
            if len(item) != 2:
                raise TypeError(f'grid indices must be ints, 2-element tuples or strings, not {item.__class__.__name__}')

            return self.location_grid[item[0], item[1]]

        if isinstance(item, str):
            index = self.grid_index_numbers(item)
            return self.location_grid[index[1]][index[0]]

        if isinstance(item, (int, slice)):
            return self.location_grid[item]

        raise TypeError(f'grid indices must be ints, slices, 2-element tuples or strings, not {item.__class__.__name__}')

    @property
    def indexes(self) -> List[str]:
        indexes = []

        for row in range(self.rows):
            for column in range(self.columns):
                indexes.append(self.grid_index(column, row))

        return indexes

    @property
    def locations(self) -> Dict[str, Location]:
        return dict(zip(self.indexes, self.location_grid.flatten()))

    def grid_index_numbers(self, grid_index: str) -> Tuple[int, int]:
        column = ord(grid_index[0].upper()) - ord('A')
        row = int(grid_index[1:]) - 1

        return column, row

    def grid_index(self, column: int, row: int) -> str:
        column_letter = 'ABCDEFGHIJKLMNOPQRSTUVQXYZ'[column]
        return f'{column_letter}{row + 1}'

    def update(self):
        pass


class StaticGrid(BaseGrid):
    def __init__(self, locations: List[Location], rows: int, columns: int):
        BaseGrid.__init__(self, rows, columns)
        self.location_grid = np.array(locations)

        if rows * columns != len(locations):
            raise StaticGridInvalidLength(f"Cannot create StaticGrid, length of locations doesn't equal rows * columns")

    def __getitem__(self, item: Union[Tuple[int, int], int, str]) -> Union[Location, np.ndarray]:
        if isinstance(item, tuple):
            if len(item) != 2:
                raise TypeError(f'grid indices must be ints, 2-element tuples or strings, not {item.__class__.__name__}')

            return self.location_grid[item[0] * self.columns + item[1]]

        if isinstance(item, str):
            index = self.grid_index_numbers(item)
            return self.location_grid[index[1] * self.columns + index[0]]

        if isinstance(item, (int, slice)):
            return self.location_grid[item]

        raise TypeError(f'grid indices must be ints, slices, 2-element tuples or strings, not {item.__class__.__name__}')


class Grid(BaseGrid):
    @classmethod
    def from_locations(cls, top_left: Location, bottom_left: Location, bottom_right: Location, rows: int, columns: int) -> 'Grid':
        x = bottom_right.position - bottom_left.position
        y = top_left.position - bottom_left.position
        z = x.cross(y)
        # use a cross product to calculate a new y vector axis relative orthogonal to x and z, then project the given
        # y vector on to it to get a new orthogonal y axis with the proper length
        y_ortho = y.project(z.cross(x))
        x_spacing = x.magnitude / (columns - 1) if columns > 1 else 0.0
        y_spacing = y_ortho.magnitude / (rows - 1) if rows > 1 else 0.0

        # location = Location.from_xy(bottom_left.position, x, y_ortho)
        spacing = Cartesian(x_spacing, y_spacing)

        return cls(bottom_left, rows=rows, columns=columns, spacing=spacing)

    def __init__(self, location: Location, rows: int, columns: int, spacing: GridSpacingType):
        BaseGrid.__init__(self, rows, columns)
        self.location = location

        if isinstance(spacing, tuple):
            self.spacing = Cartesian(*spacing)
        elif isinstance(spacing, Cartesian):
            self.spacing = spacing
        else:
            raise GridSpacingError(f'Invalid grid spacing, must be a 2- or 3-element tuple or Cartesian: {spacing}')

        self.update()

    def update(self):
        base_grid = np.array([[Location(x, y, 0) for x in range(self.columns)] for y in range(self.rows)])
        self.location_grid = self.location * self.spacing.scale(base_grid)


class GridError(Exception):
    pass


class GridSpacingError(GridError):
    pass


class StaticGridError(Exception):
    pass


class StaticGridInvalidLength(StaticGridError):
    pass



from typing import List, Tuple, Dict, Optional, Union, Any
import json
from hein_robots.base.robot_arms import RobotArm
from hein_robots.grids import LocationGroup
from hein_robots.robotics import Location


class ManagerError(Exception):
    pass


class LocationManagerError(ManagerError):
    pass


class LocationManagerEmptyError(LocationManagerError):
    pass


class LocationManager:
    """
    The LocationManager class is a simple class that is meant to be subclassed for working with specific location groups.
    LocationManager classes will keep track of the "current" index in a list of grids, and allow the user to increment the
    "current" index to the next location. It will also raise a LocationManagerEmptyError once it has run out of indexes.
    """
    def __init__(self, robot: RobotArm, groups: List[LocationGroup], state_file: Optional[str]=None):
        """
        :param robot: an instance of RobotArm
        :param groups: a list of LocationGroup instances
        """
        self.robot = robot
        self.groups = groups
        self.current_indexes_index = 0
        self.indexes: List[Tuple[LocationGroup, str]] = []
        self.state_file = state_file

        self.indexes = self.build_indexes()
        self.load_state()

    @property
    def remaining(self) -> int:
        """ Get the number of remaining indexes """
        return len(self.indexes) - self.current_indexes_index

    @property
    def empty(self) -> bool:
        """ Return True if there are no remaining indexes """
        return self.current_indexes_index >= len(self.indexes)

    @property
    def current_location(self) -> Location:
        """ Get the location at the current index """
        group, index = self.indexes[self.current_indexes_index]
        return group.locations[index]

    @property
    def current_group(self) -> LocationGroup:
        """ Get the current LocationGroup instance """
        group, index = self.indexes[self.current_indexes_index]
        return group

    @property
    def current_index(self) -> str:
        """ Get the current index key """
        group, index = self.indexes[self.current_indexes_index]
        return index

    @property
    def state(self) -> Dict[str, Any]:
        """ Get or set a dictionary with the current state of the LocationManager """
        return {
            'current_indexes_index': self.current_indexes_index
        }

    def reset(self):
        """ Reset the current index to the start """
        self.current_indexes_index = 0
        self.save_state()

    def build_indexes(self):
        indexes = []

        for group in self.groups:
            indexes += [(group, index) for index in group.indexes()]

        return indexes

    def increment_location(self):
        """ Incrememnt the current index, raises a LocationManagerEmptyError if empty """
        if self.remaining <= 0:
            raise LocationManagerEmptyError(f'Location groups empty')

        self.current_indexes_index += 1
        self.save_state()

    def save_state(self):
        if self.state_file is None:
            return

        with open(self.state_file, 'w') as state_file:
            json.dump(self.state, state_file)

    def load_state(self):
        if self.state_file is None:
            return

        try:
            with open(self.state_file, 'r') as state_file:
                state = json.load(state_file)

                for key, val in state.items():
                    setattr(self, key, val)
        except FileNotFoundError:
            pass



from typing import Union
from hein_robots.base.actuators import Actuator


class Gripper(Actuator):
    @property
    def is_open(self) -> bool:
        return self.position < 0.5

    @property
    def is_closed(self) -> bool:
        return not self.is_open

    def open(self, position: Union[float, bool] = True):
        if isinstance(position, bool):
            return self.move(1.0 if position else 0.0)

        self.move(position)

    def close(self, position: Union[float, bool] = False):
        if isinstance(position, bool):
            return self.open(not position)

        self.open(position)
